import java.io.*;
import java.util.*;
import jcifs.smb.*;
import java.net.URI;
import jcifs.Config;
import javax.naming.*;
import javax.naming.ldap.*;
import java.io.IOException;
import javax.naming.ldap.*;
import java.net.URLEncoder;
import jcifs.util.LogStream;
import javax.naming.Context;
import javax.naming.directory.*;
import java.net.URISyntaxException;
import javax.naming.NamingException;
import javax.naming.NamingEnumeration;

class cifsls
{
    private static final String VERSION = "1.25";

    private static final String umess =
        "Usage: java cifsls [-l] [-d] [-w] [-ld <ldap.cfg>] <cifsVolume>\n" +
        "\tcifsVolume should be of the form\n" +
        "\t    //<host>/<share>\n" +
        "\tor  //<host>/<share>/path\n\n" +
        "  -d  disable DFS\n" +
        "  -w  disable IG WAR\n" +
        "  -l  list directory contents\n" +
        "  -ld specify ldap config file\n";

    private static final String ldcfmess =
        "LDAP config file format is a plain text file with two lines.\n" +
        "First line should be the URL for the AD server.\n" +
        "    (e.g. ldap://ad-host.corp.iongrid.com:389)\n" +
        "Second line should be the base DN.\n" +
        "(You can find this in the Nexus dashboard, by under the admin\n" +
        " page, in the Active Directory section.)\n";

    private boolean disableDFS, disableChecking, listFiles;
    private String volName;

    cifsls(String[] arghv)
    {
        listFiles = disableDFS = false;
        volName = null;
        for (int i = 0; i < arghv.length; i++) {
            String a = arghv[i];
            if ("-d".equals(a)) disableDFS = true;
            else if ("-l".equals(a)) listFiles = true;
            else if ("-w".equals(a)) disableChecking = true;
            else if ("-ld".equals(a)) {
                if (i + 1 == arghv.length) usage(umess);
                configLDAP(arghv[++i]);
            } else if (volName == null) volName = a;
            else usage(umess);
        }

        if (volName == null) usage(umess);
    }

    private static void usage(String mess)
    {
        System.out.println(mess);
        System.exit(1);
    }

    private static final String makeSmbUrl(String s)
    {
        if (s == null) return s;

        s = s.replace('\\', '/');
        if (!s.startsWith("smb:")) s = "smb:" + s;
        if (!s.endsWith("/")) s = s + '/';

        return s;
    }

    private NtlmPasswordAuthentication getAuth()
        throws Exception
    {
        Console con = System.console();
        String user, pass;

        if (con == null) {
            // throw new IOException("No console found.");
            user = "ross";
            pass = "Hlay,k";
        } else {
            user = con.readLine("NT domain: ") + "\\" + con.readLine("username: ");
            pass = new String(con.readPassword("password: "));
        }

        volName = translateDFSHost(volName);
        return new NtlmPasswordAuthentication(null, user, pass);
    }

    private void configLDAP(String ldfname)
    {
        try {
            BufferedReader rdr = new BufferedReader(new FileReader(ldfname));
            String url = rdr.readLine();
            String baseDN = rdr.readLine().trim();
            if (baseDN == null || (baseDN = baseDN.trim()).isEmpty() ||
                (url = url.trim()).isEmpty() ||
                rdr.readLine() != null)
                usage(ldcfmess);
            Console con = System.console();
            String pass;
            if (con == null)
                pass = "pass4adquery";
                // throw new IOException("No console found.");
            else pass = new String(con.readPassword("AD password: "));
            configLDAP(url, baseDN, pass);
        } catch (Exception e) {
            System.out.println(e.getMessage());
            usage(ldcfmess);
        }
    }

    private static File logFileName()
        throws IOException
    {
        File ans;
        int ctr;

        ans = new File("cifs.log");
        for (ctr = 1; true; ctr++) {
            if (!ans.exists()) return ans;
            ans = new File(String.format("cifs%04d.log", ctr));
        }
    }

    static PrintStream log;

    private static void spew(String s, Object... args)
    {
        if (args.length > 0)
            s = String.format(s, args);
        System.out.println(s);
        if (log != null)
            log.println(s);
    }

    public static void main(String[] arghv)
        throws Exception
    {
        LogStream.setInstance(new PrintStream(logFileName()));
        LogStream.setLevel(25);
        log = LogStream.getInstance();
        spew("CIFS analyzer tool " + VERSION);

        cifsls me = new cifsls(arghv);
        if (me.disableDFS)
            Config.setProperty("jcifs.smb.client.dfs.disabled", "true");
        if (me.disableChecking)
            Config.setProperty("jcifs.smb.client.dfs.watch", "false");

        try {
            NtlmPasswordAuthentication auth = me.getAuth();
            long start = System.currentTimeMillis();
            String volume = makeSmbUrl(me.volName);
            SmbFile rt = new SmbFile(volume, auth);

            spew("%s does%s exist", volume, rt.exists() ? "" : "n't");
            spew("%s is%s a directory", volume, rt.isDirectory() ? "" : "n't");
            spew("%s can%s be read", volume, rt.canRead() ? "" : "'t");
            int nd = 0, nf = 0;
            for (SmbFile f : rt.listFiles()) {
                boolean isdir = f.isDirectory();
                if (isdir) nd++;
                else nf++;
                if (me.listFiles)
                    spew("[%s] %s", isdir ? "d" : "f", f);
            }
            long end = System.currentTimeMillis();
            spew("ls() yields %d files and %d dirs", nf, nd);
            spew("runtime: %.3f seconds", (end - start) / 1000.);
        } catch (Throwable t) {
            spew("Uncaught exception.");
            t.printStackTrace();
            t.printStackTrace(log);
        }
    }

    /* ---------------------------------------- */

    protected static final String LDAP_CONTEXT_FACTORY =
        "com.sun.jndi.ldap.LdapCtxFactory";

    private static String configurationNamingContext = null;
    private static Hashtable<String,String> _env;

    static void configLDAP(String url, String baseDN, String adPass)
    {
        _env = new Hashtable<String,String>();

        _env.put(Context.INITIAL_CONTEXT_FACTORY, LDAP_CONTEXT_FACTORY);
        _env.put(Context.PROVIDER_URL, url);
        _env.put(Context.SECURITY_PRINCIPAL, baseDN);
        _env.put(Context.SECURITY_CREDENTIALS, adPass);
        spew("LDAP config: URL = %s", url);
    }

    private String getCfgNamingCtxt() throws NamingException
    {
        InitialLdapContext ctx = new InitialLdapContext(_env, null);
        try {
            String attr = "configurationNamingContext";
            return (String)
                ctx.getAttributes("", new String[] { attr }).get(attr).get();
        } catch (NamingException ne) {
            return null;
        } finally {
            try { ctx.close(); } catch (Exception ex) {}
        }
    }

    private String getFQDN(String biosName) throws NamingException
    {
        Query q = new Query();

        q.base = String.format("CN=Partitions,%s", getCfgNamingCtxt());
        q.filter = String.format("(nETBIOSName=%s)", biosName);
        q.constraints.setCountLimit(1);
        q.constraints.setReturningAttributes(new String[] {"dnsRoot"});
        q.constraints.setSearchScope(SearchControls.SUBTREE_SCOPE);

        Query.Result r = q.execute();
        if (!r.hasMore()) {
            spew("NETbios->FQDN translate fail: %s => null", biosName);
            return null;
        }
        return r.next()[0];
    }

    String translateDFSHost(String in)
    {
        int b, e;
        char chrs[];
        String host;

        if (in == null || _env == null) return in;
        chrs = new char[in.length()];
        in.getChars(0, in.length(), chrs, 0);
        for (b = 0;
             b < chrs.length && (chrs[b] == '\\' || chrs[b] == '/'); b++);
        for (e = b;
             e < chrs.length && chrs[e] != '\\' && chrs[e] != '/'; e++);
        try {
            host = getFQDN(new String(chrs, b, e - b));
        } catch (Exception ex) {
            spew("translate fail: %s", ex.getMessage());
            host = null;
        }
        if (host != null) {
            spew("translateHost old: %s", in);
            in = in.substring(0, b) + host + in.substring(e);
            spew("translateHost new: %s", in);
        }
        return in;
    }

    protected static Map<String,String> mapList(final String... args) {
        return new HashMap<String, String>(){{
                for (int i = 0; i < args.length; i+=2) put(args[i],args[i+1]);
            }};
    }

    //
    // query abstraction
    //
    static class Query
    {
        public String base;
        public String filter;
        public SearchControls constraints = new SearchControls();
        public PagedResultsControl paging = null;

        public Query() { }

        public Query(Query model) {
            base = model.base;
            filter = model.filter;
            constraints.setCountLimit(model.constraints.getCountLimit());
            constraints.setDerefLinkFlag(model.constraints.getDerefLinkFlag());
            constraints.setReturningAttributes(model.constraints.getReturningAttributes());
            constraints.setReturningObjFlag(model.constraints.getReturningObjFlag());
            constraints.setSearchScope(model.constraints.getSearchScope());
            constraints.setTimeLimit(model.constraints.getTimeLimit());
        }

        Result _execute(String base, String filter, SearchControls constraints)
            throws NamingException
        {
            InitialLdapContext ctx = null;
            try {
                ctx = new InitialLdapContext(_env, null);
                if (paging != null)
                    ctx.setRequestControls(new Control[] { paging });
                NamingEnumeration e = ctx.search(base, filter, constraints);
                Control[] controls = ctx.getResponseControls();
                return new Result(e, controls);
            } finally {
                try {
                    if (ctx != null) ctx.close();
                }
                catch (Exception e) {
                    System.err.println("Exception closing connection" + e);
                }
            }
        }

        public Result execute() throws NamingException {
            return _execute(base, filter, constraints);
        }

        public Result execute(String... substitutes) throws NamingException
        {
            return execute(mapList(substitutes));
        }

        public Result execute(Map<String,String> substitutes) throws NamingException {
            String subFilter = filter;
            for (Map.Entry<String,String> kvp : substitutes.entrySet()) {
                // java replaceAll is broken - it may fail depending
                //  on the contents of subFilter see
                //  http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4689750
                //  and they insist its not a bug...
                subFilter = subFilter.replace(kvp.getKey(),kvp.getValue());
            }
            return _execute(base, subFilter, constraints);
        }

        static String join(String delim, String elts[])
        {
            StringBuilder ans = new StringBuilder();
            for (String s : elts) {
                if (ans.length() > 0) ans.append(delim);
                ans.append(s);
            }

            return ans.toString();
        }

        @Override public String toString() {
            return String.format(
                    "LdapQuery: base=%s,filter=%s,scope=%d,retattrs=%s",
                    base,
                    filter,
                    constraints.getSearchScope(),
                    join(",", constraints.getReturningAttributes())
            );
        }

        // query result abstraction that turns turns attributes into lists
        public class Result implements NamingEnumeration<String[]> {
            public NamingEnumeration ul;
            public byte[] cookie;

            public Result(NamingEnumeration underlying) { ul = underlying; }
            public Result(NamingEnumeration underlying, Control[] controls)
            {
                ul = underlying;
                if (controls != null) {
                    for (Control c : controls) {
                        if (c instanceof PagedResultsResponseControl)
                            cookie = ((PagedResultsResponseControl)c).getCookie();
                    }
                }
            }

            public String[] next() throws NamingException {
                if (!hasMoreElements()) return null;
                SearchResult sr = (SearchResult)ul.next();
                Attributes attrs = sr.getAttributes();
                String retAttr[] =
                    Query.this.constraints.getReturningAttributes();
                String ans[] = new String[retAttr.length];
                int i;

                if (retAttr != null) {
                    for (i = 0; i < retAttr.length; i++) {
                        if ("distinguishedName".equalsIgnoreCase(retAttr[i])) {
                            try {
                                ans[i] = sr.getNameInNamespace();
                            } catch (UnsupportedOperationException ex) {
                                throw new NamingException("Unable to resolve object DN");
                            }
                        }
                    }
                }

                if (attrs != null && retAttr != null) {
                    for (NamingEnumeration e = attrs.getAll(); e.hasMore();) {
                        Attribute a = (Attribute)e.next();
                        if (a != null) {
                            for (i = 0; i < retAttr.length &&
                                     !a.getID().equalsIgnoreCase(retAttr[i]); i++);
                            if (i < ans.length && !"distinguishedName".equalsIgnoreCase(retAttr[i]))
                                ans[i] = (String) a.get(0);
                        }
                    }
                }
                return ans;
            }

            public boolean hasMore() throws NamingException { return ul.hasMore(); }
            public void close() throws NamingException { ul.close(); }

            public boolean hasMoreElements() { return ul.hasMoreElements(); }
            public String[] nextElement()
            {
                try { return next(); } catch (NamingException e) { return null; }
            }
        }
    }
}
